/*
 * Copyright (C) 1997 Michael Davidson.
 * All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software
 * and its documentation for any purpose and without fee is hereby
 * granted, provided that the above copyright notice appear in all
 * copies and that both that copyright notice and this permission
 * notice appear in supporting documentation.
 *
 * This software is provided "as is" without express or implied warranty.
 */

#pragma ident	"@(#)pathname.c	1.5	99/02/25 SMI"

#include	<errno.h>
#include	<string.h>
#include	<stdlib.h>
#include	<stdio.h>
#include	<ctype.h>
#include	<sys/types.h>
#include	<sys/wait.h>
#include	<sys/stat.h>
#include	<fcntl.h>
#include	<unistd.h>
#include	"lxrun.h"

extern char	*linux_root;


#ifdef PATHMAPFILE

/* The pathname remapping table.  S.G. */

typedef struct pathmap_T {
	enum { EXACT, PREFIX }	type;		/* What type of match? */
	char			*from, *to;	/* Map from & to strings */
	char			toisexecutable;	/* Is "to" an executable
						** that generates a file's
						** text on-the-fly?
						*/
	struct pathmap_T	*next;		/* Next mapping in list */
} pathmap_t;

pathmap_t		*pathmap_head	= NULL;
char			phantom_file[20];
char			phantom_file_valid = 0;



/*
** Create a phantom file based on the information in pm.
** Return its name in fn, unless it's longer than len.
** If the name is too long, return -1 and set errno to ENAMETOOLONG.
*/
int create_phantom_file(pathmap_t *pm, const char *from, char *fn, int len)
{
	int fd, phlen;
	char fd_str[10];

	/*
	** It's okay that the name is always the same for this process
	** since we delete it immediately anyway.  If we need more than
	** one phantom file per system call, then we're in trouble, and
	** this mechanism will need to be changed.
	*/
	sprintf(phantom_file, "/tmp/lx%d", getpid());

	/*
	fprintf(stderr, "lxrun: Calling %s to emulate %s\n", pm->to, from);
	*/

	if((phlen = strlen(phantom_file))+1 > len)
		{
		errno = ENAMETOOLONG;
		return -1;
		}

	/* Create the phantom file */
	if((fd=open(phantom_file, O_RDWR | O_CREAT, 0600))==-1)
		return -1;
	phantom_file_valid=1;
	sprintf(fd_str, "%d", fd);

	/* Send the name of the phantom file back to the Linux process */
	strcpy(fn, phantom_file);

	/* Call our executable to write out the file */
	if(fork() == 0)
		{
		char buf[512];

		/* cmdline; truncate if necessary */
		strcpy(buf, "LINUX_CMDLINE=");
		strncpy(buf+14, linux_cmdline, (sizeof buf)-15);
		buf[sizeof buf - 1] = '\0';
		putenv(strdup(buf));

		/* our pid */
		sprintf(buf, "LINUX_PID=%d", getpid());
		putenv(strdup(buf));

		/*
		** Write the phantom file represeting "from" to
		** file descriptor "fd".
		*/
		execl(pm->to, pm->to, fd_str, from, NULL);
		/* We should print out an error message here, maybe. */
		exit(-1);
		}

	/* The parent doesn't need the phantom file fd */
	close(fd);

	/* Wait for the child to write the phantom file. */
	wait((int *)0);

	return phlen;
}


/* Remove any phantom file */
void remove_phantom_file(void)
{
	if(phantom_file_valid) {
		unlink(phantom_file);
		phantom_file_valid = 0;
	}
}

#endif	/* PATHMAPFILE */


/*
 * safe getstring() for fetching pathnames from the
 * program being run under emulation - checks that
 * each page is readable before copying
 * returns total number of bytes copied (including the
 * null terminator) or -1 in the case of a fault
 */

int
getstring(const void *src, void *dst, int len)
{
	register const char *s = (const char *)src;
	register       char *d = (char *)dst;

	while (len > 0)
	{
		register int n;

		if (getbyte(s) < 0)	/* check for faults */
			return -1;

		n    = PAGE_SIZE - ((unsigned long)src & (PAGE_SIZE-1));
		len -= n;

		while (--n >= 0)
			if (! (*d++ = *s++))
				return d - (char *)dst;
	}
	
	return d - (char *)dst;
}

/*
 * map Linux path names into system path names
 * (mainly for the benefit of the dynamic linker
 * when attempting to load shared libraries)
 */
int
lx_pathname(const char *lx_path, char *os_path, int len)
{
	int		n = getstring(lx_path, os_path, len);
	int		r;
#ifdef PATHMAPFILE
	pathmap_t	*pm;
#endif

	if (n < 0) {
		errno = EFAULT;
		return -1;
	}

	if (n == len && os_path[n-1] != '\0') {
		errno = ENAMETOOLONG;
		return -1;
	}

#ifdef PATHMAPFILE
	/* If the input name starts with the root simply return. */
	{
		size_t len = strlen(linux_root);
		
		if (strncmp(lx_path, linux_root, len) == 0) {
			(void) strcpy(os_path, lx_path);
			return len;
		}
	}
			
	/*
	** Traverse linked list of pathname maps and apply the first
	** appropriate one we find.
	*/
	for(pm = pathmap_head; pm; pm=pm->next)
		{
		if(pm->type == EXACT && strcmp(lx_path, pm->from) == 0)
			{
			if(pm->toisexecutable)
				return create_phantom_file(pm,
						lx_path, os_path, len);
			strncpy(os_path, pm->to, len);
			return strlen(os_path);
			}

		else if(pm->type == PREFIX &&
				strncmp(lx_path, pm->from, strlen(pm->from))==0)
			{
			if(pm->toisexecutable)
				return create_phantom_file(pm,
						lx_path, os_path, len);
			if(len < (int)(strlen(lx_path) - strlen(pm->from) +
					strlen(pm->to) + 1))
				{
				errno = ENAMETOOLONG;
				return -1;
				}
			strcpy(os_path, pm->to);
			strcat(os_path, lx_path+strlen(pm->from));
			return strlen(os_path);
			}
		}

	/* No matches were made.  Return the input as-is. */
	strcpy(os_path, lx_path);

	return strlen(os_path);
#else

	/*
	 * simple hack for now - just trap accesses to /lib and /usr/lib
	 */

	if (linux_root && (strncmp(os_path, "/lib/", 5) == 0 ||
		strncmp(os_path, "/usr/lib/", 9) == 0 ||
		strncmp(os_path, "/usr/local/lib/", 15) == 0 ||
		strncmp(os_path, "/usr/X11R6/lib/", 15) == 0) ) {

		r = strlen(linux_root);
		if (r + n > len) {
			errno = ENAMETOOLONG;
			return -1;
		}
		memmove(os_path+r, os_path, n);
		memcpy(os_path, linux_root, r);

		return n + r;
	}

	return n;

#endif	/* PATHMAPFILE */
}


#ifdef PATHMAPFILE

int parse_pathmap_table(const char *fn)
{
	FILE *fp;
	char buf[256], *s, *from, *to;
	int linenum;
	pathmap_t *newent;

	/* open pathmap definition file */
	if((fp=fopen(fn, "r"))==NULL)
		return -1;

	for(linenum=1; (s=fgets(buf, sizeof buf, fp)) != NULL; linenum++)
		{
		while(isspace(*s) && *s != '\0')
			s++;		/* skip leading whitespace */

		if(*s == '#' || *s == ';')
			continue;	/* comments */

		/*
		** assert: *s is the first non-whitespace character in
		** a non-comment line.
		*/
		if(*s == '\0')		/* empty line */
			continue;

		from=s;

		while(!isspace(*s) && *s != '\0')
			s++;

		/*
		** assert: *s is the first whitespace character (or nul)
		** after the "from" field.
		*/
		while(isspace(*s) && *s != '\0')
			*s++ = '\0';

		if(*s == '\0')		/* no "to" field! */
			{
			warning(0, "parse error in %s line %d\n", fn, linenum);
			continue;
			}

		/* assert: *s is the first character in the "to" field. */
		to=s;

		while(!isspace(*s) && *s != '\0')
			s++;

		*s='\0';

		/*
		** If we made it this far, then from and to point to
		** nul-terminated strings each containing at least one
		** character and neither containing any whitespace.
		*/

		/* Note: this memory is never deallocated. */
		newent = (pathmap_t *)malloc(sizeof(pathmap_t));

		if(from[strlen(from)-1] == '*')
			{
			newent->type = PREFIX;
			from[strlen(from)-1] = '\0';
			}
		else
			newent->type = EXACT;

		newent->from = strdup(from);

		newent->toisexecutable = 0;

		if(to[0]=='*' && to[1]=='\0')
			{
			newent->to = (char *)malloc(
					strlen(linux_root) + strlen(from) + 1);
			strcpy(newent->to, linux_root);
			strcat(newent->to, from);
			}
		else if(to[0]=='@')	/* executable that emulates a file */
			{
			newent->toisexecutable = 1;
			newent->to = strdup(to+1);
			}
		else
			newent->to = strdup(to);

		/* Insert newent into the linked list. */
		newent->next = pathmap_head;
		pathmap_head = newent;
		}

	fclose(fp);

	return 0;
}

#endif

void
find_emulator(char *pathenv, char *emulator, const char *emul)
{
	struct stat buf;

	/* see if the passed in name is right */
	if (stat(emul, &buf) == 0) {
		if (S_ISREG(buf.st_mode) && (access(emul, X_OK) == 0)) {
			strcpy(emulator, emul);
			return;
		}
	}

	if (pathenv != NULL) {	/* Check for lxrun binary in PATH */
		/* We use a copy of pathenv since strtok is destructive. */
		char *pathcopy = strdup(pathenv);

		char *pathp = pathcopy;

		while ((pathp=strtok(pathp, ":")) != NULL) {
			strcpy(emulator, pathp);
			strcat(emulator, "/");
			strcat(emulator, emul);

			if (stat(emulator, &buf) == 0) {
				if (S_ISREG(buf.st_mode) &&
				    (access(emulator, X_OK) == 0)) {
					free(pathcopy);
					return;
				}
			}
			pathp=NULL;
		}
		free(pathcopy);
	}
	fprintf(stderr, "%s: Can't find full path of lxrun binary!", emul);
	exit (1);
}
